/*	$Id: tree.c,v 1.36 2011/02/09 09:18:15 kristaps Exp $ */
/*
 * Copyright (c) 2011 Kristaps Dzonsons <kristaps@bsd.lv>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>

#include <assert.h>
#include <db.h>
#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mandoc.h"
#include "main.h"

enum	indexflags {
	DBNAME		= 0x01, /* manual name(s) (NAME, Nm) */
	DBUTIL		= 0x02, /* utility (SYNOPSIS, Nm) */
	DBFUNC		= 0x04, /* function (SYNOPSIS, Fn, Fo) */
};

struct	index {
	DB		*db; /* btree database */
	char		*curbuf; /* buffer of flag,filename */
	size_t	 	 curbufsz; /* size of curbuf */
	const char	*file; /* database file */
};

typedef	void		(*mdoc_indexh)(struct index *, const struct mdoc_node *);
static	void	 	 scan_mdoc(struct index *, const struct mdoc_node *);
static	void	 	 scan_fn(struct index *, const struct mdoc_node *);
static	void	 	 scan_fo(struct index *, const struct mdoc_node *);
static	void	 	 scan_nm(struct index *, const struct mdoc_node *);
static	void	 	 scan_put(struct index *, char *, int);

static	mdoc_indexh mdocs[MDOC_MAX] = {
	NULL, /* Ap */
	NULL, /* Dd */
	NULL, /* Dt */
	NULL, /* Os */
	NULL, /* Sh */
	NULL, /* Ss */ 
	NULL, /* Pp */ 
	NULL, /* D1 */
	NULL, /* Dl */
	NULL, /* Bd */
	NULL, /* Ed */
	NULL, /* Bl */
	NULL, /* El */
	NULL, /* It */
	NULL, /* Ad */ 
	NULL, /* An */
	NULL, /* Ar */
	NULL, /* Cd */
	NULL, /* Cm */
	NULL, /* Dv */ 
	NULL, /* Er */ 
	NULL, /* Ev */ 
	NULL, /* Ex */
	NULL, /* Fa */ 
	NULL, /* Fd */ 
	NULL, /* Fl */
	scan_fn, /* Fn */ 
	NULL, /* Ft */ 
	NULL, /* Ic */ 
	NULL, /* In */ 
	NULL, /* Li */
	NULL, /* Nd */ 
	scan_nm, /* Nm */ 
	NULL, /* Op */
	NULL, /* Ot */
	NULL, /* Pa */
	NULL, /* Rv */
	NULL, /* St */ 
	NULL, /* Va */
	NULL, /* Vt */ 
	NULL, /* Xr */
	NULL, /* %A */
	NULL, /* %B */
	NULL, /* %D */
	NULL, /* %I */
	NULL, /* %J */
	NULL, /* %N */
	NULL, /* %O */
	NULL, /* %P */
	NULL, /* %R */
	NULL, /* %T */
	NULL, /* %V */
	NULL, /* Ac */
	NULL, /* Ao */
	NULL, /* Aq */
	NULL, /* At */
	NULL, /* Bc */
	NULL, /* Bf */ 
	NULL, /* Bo */
	NULL, /* Bq */
	NULL, /* Bsx */
	NULL, /* Bx */
	NULL, /* Db */
	NULL, /* Dc */
	NULL, /* Do */
	NULL, /* Dq */
	NULL, /* Ec */
	NULL, /* Ef */
	NULL, /* Em */ 
	NULL, /* Eo */
	NULL, /* Fx */
	NULL, /* Ms */
	NULL, /* No */
	NULL, /* Ns */
	NULL, /* Nx */
	NULL, /* Ox */
	NULL, /* Pc */
	NULL, /* Pf */
	NULL, /* Po */
	NULL, /* Pq */
	NULL, /* Qc */
	NULL, /* Ql */
	NULL, /* Qo */
	NULL, /* Qq */
	NULL, /* Re */
	NULL, /* Rs */
	NULL, /* Sc */
	NULL, /* So */
	NULL, /* Sq */
	NULL, /* Sm */ 
	NULL, /* Sx */
	NULL, /* Sy */
	NULL, /* Tn */
	NULL, /* Ux */
	NULL, /* Xc */
	NULL, /* Xo */
	scan_fo, /* Fo */ 
	NULL, /* Fc */ 
	NULL, /* Oo */
	NULL, /* Oc */
	NULL, /* Bk */
	NULL, /* Ek */
	NULL, /* Bt */
	NULL, /* Hf */
	NULL, /* Fr */
	NULL, /* Ud */
	NULL, /* Lb */
	NULL, /* Lp */ 
	NULL, /* Lk */ 
	NULL, /* Mt */ 
	NULL, /* Brq */ 
	NULL, /* Bro */ 
	NULL, /* Brc */ 
	NULL, /* %C */ 
	NULL, /* Es */ 
	NULL, /* En */
	NULL, /* Dx */ 
	NULL, /* %Q */ 
	NULL, /* br */
	NULL, /* sp */ 
	NULL, /* %U */ 
	NULL, /* Ta */ 
};

/* ARGSUSED */
void
index_man(void *arg, const struct man *m)
{

	/* Do nothing. */
}

void *
index_alloc(char *arg)
{
	struct index	*db;
	const char	*file;
	BTREEINFO	 info;

	db = mandoc_calloc(1, sizeof(struct index));

	memset(&info, 0, sizeof(BTREEINFO));
	info.flags = R_DUP;

	db->file = "man.btree";
	db->db = dbopen(db->file, O_CREAT | O_RDWR, 0644, DB_BTREE, &info);

	if (NULL == db->db) {
		perror(file);
		exit((int)MANDOCLEVEL_SYSERR);
	}

	return(db);
}

void
index_free(void *arg)
{
	struct index	*db;

	db = (struct index *)arg;

	if (-1 == (*db->db->close)(db->db))
		perror(db->file);
	if (db->curbuf)
		free(db->curbuf);

	free(db);
}

/* ARGSUSED */
void
index_mdoc(void *arg, const struct mdoc *m)
{
	struct index	*db;
	const char	*file;
	size_t		 filesz;

	db = (struct index *)arg;

	if (NULL == (file = mdoc_meta(m)->file))
		return;

	/*
	 * Create enough storage space for the record type and the
	 * file-name of the record.
	 */

	filesz = strlen(file);

	db->curbufsz = filesz + 5; /* Four bytes (int) and nil. */
	db->curbuf = mandoc_realloc(db->curbuf, db->curbufsz);

	strlcpy(db->curbuf + 4, file, db->curbufsz - 4);
	scan_mdoc(db, mdoc_node(m));
}

static void
scan_put(struct index *db, char *buf, int flag)
{
	DBT		 key, val;

	if ('\0' == buf[0])
		return;

	key.data = buf;
	key.size = strlen(buf);

	/*
	 * The value of the record is the entry type (host-byte integer
	 * bit-field) followed by the nil-terminated filename.
	 */

	memcpy(db->curbuf, &flag, 4);

	val.data = db->curbuf;
	val.size = db->curbufsz;

	if (-1 == (*db->db->put)(db->db, &key, &val, 0)) {
		perror(db->file);
		exit((int)MANDOCLEVEL_SYSERR);
	}
}

/*
 * Accept function names (`Fn') in the SYNOPSIS section.  `Fn' has
 * strange syntax, so make sure we get the actual name and not the type.
 */
static void
scan_fn(struct index *db, const struct mdoc_node *n)
{

	if (SEC_SYNOPSIS != n->sec)
		return;
	if (MDOC_ELEM != n->type)
		return;

	if (n->nchild > 1) 
		n = n->child->next;
	else
		n = n->child;

	scan_put(db, n->string, DBFUNC);
}

static void
scan_fo(struct index *db, const struct mdoc_node *n)
{
	char		 buf[BUFSIZ];

	if (SEC_SYNOPSIS != n->sec)
		return;
	if (MDOC_ELEM != n->type)
		return;

	for (buf[0] = '\0', n = n->child; n; n = n->next) {
		assert(MDOC_TEXT == n->type);
		strlcat(buf, n->string, BUFSIZ);
		if (n->next)
			strlcat(buf, " ", BUFSIZ);
	}

	scan_put(db, buf, DBFUNC);
}

static void
scan_nm(struct index *db, const struct mdoc_node *n)
{
	int		 flag;
	char		 buf[BUFSIZ];

	if (SEC_NAME != n->sec && SEC_SYNOPSIS != n->sec)
		return;
	if (MDOC_ELEM != n->type && MDOC_HEAD != n->type)
		return;

	flag = SEC_NAME == n->sec ? DBNAME : DBUTIL;

	for (buf[0] = '\0', n = n->child; n; n = n->next) {
		if (MDOC_TEXT != n->type)
			continue;
		strlcat(buf, n->string, BUFSIZ);
		if (n->next)
			strlcat(buf, " ", BUFSIZ);
	}

	scan_put(db, buf, flag);
}

static void
scan_mdoc(struct index *db, const struct mdoc_node *n)
{

	switch (n->type) {
	case (MDOC_ELEM):
		/* FALLTHROUGH */
	case (MDOC_BLOCK):
		/* FALLTHROUGH */
	case (MDOC_TAIL):
		/* FALLTHROUGH */
	case (MDOC_BODY):
		/* FALLTHROUGH */
	case (MDOC_HEAD):
		if (NULL == mdocs[(int)n->tok])
			break;
		(*mdocs[(int)n->tok])(db, n);
		break;
	default:
		break;
	}

	if (n->child)
		scan_mdoc(db, n->child);
	if (n->next)
		scan_mdoc(db, n->next);
}

